@@ -8,7 +8,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclUtf.c,v 1.28 2002/08/05 03:24:41 dgp Exp $
+ * RCS: @(#) $Id: tclUtf.c,v 1.30.2.2 2003/10/08 14:21:20 dkf Exp $
  */
 
 #include "tclInt.h"
@@ -423,7 +423,7 @@ Tcl_UtfToUniCharDString(string, length, dsPtr)
     w = wString;
     end = string + length;
     for (p = string; p < end; ) {
-	p += Tcl_UtfToUniChar(p, w);
+	p += TclUtfToUniChar(p, w);
 	w++;
     }
     *w = '\0';
@@ -490,27 +490,33 @@ Tcl_NumUtfChars(str, len)
 {
     Tcl_UniChar ch;
     register Tcl_UniChar *chPtr = &ch;
-    register int n;
-    int i;
+    register int i;
 
     /*
      * The separate implementations are faster.
+     *
+     * Since this is a time-sensitive function, we also do the check for
+     * the single-byte char case specially.
      */
 
     i = 0;
     if (len < 0) {
-	while (1) {
-	    str += Tcl_UtfToUniChar(str, chPtr);
-	    if (ch == '\0') {
-		break;
-	    }
+	while (*str != '\0') {
+	    str += TclUtfToUniChar(str, chPtr);
 	    i++;
 	}
     } else {
+	register int n;
+
 	while (len > 0) {
-	    n = Tcl_UtfToUniChar(str, chPtr);
-	    len -= n;
-	    str += n;
+	    if (UCHAR(*str) < 0xC0) {
+		len--;
+		str++;
+	    } else {
+		n = Tcl_UtfToUniChar(str, chPtr);
+		len -= n;
+		str += n;
+	    }
 	    i++;
 	}
     }
@@ -545,7 +551,7 @@ Tcl_UtfFindFirst(string, ch)
     Tcl_UniChar find;
 
     while (1) {
-	len = Tcl_UtfToUniChar(string, &find);
+	len = TclUtfToUniChar(string, &find);
 	if (find == ch) {
 	    return string;
 	}
@@ -587,7 +593,7 @@ Tcl_UtfFindLast(string, ch)
 
     last = NULL;
     while (1) {
-	len = Tcl_UtfToUniChar(string, &find);
+	len = TclUtfToUniChar(string, &find);
 	if (find == ch) {
 	    last = string;
 	}
@@ -625,7 +631,7 @@ Tcl_UtfNext(str)
 {
     Tcl_UniChar ch;
 
-    return str + Tcl_UtfToUniChar(str, &ch);
+    return str + TclUtfToUniChar(str, &ch);
 }
 
 /*
@@ -706,7 +712,7 @@ Tcl_UniCharAtIndex(src, index)
 
     while (index >= 0) {
 	index--;
-	src += Tcl_UtfToUniChar(src, &ch);
+	src += TclUtfToUniChar(src, &ch);
     }
     return ch;
 }
@@ -737,7 +743,7 @@ Tcl_UtfAtIndex(src, index)
 
     while (index > 0) {
 	index--;
-	src += Tcl_UtfToUniChar(src, &ch);
+	src += TclUtfToUniChar(src, &ch);
     }
     return src;
 }
@@ -825,7 +831,7 @@ Tcl_UtfToUpper(str)
 
     src = dst = str;
     while (*src) {
-        bytes = Tcl_UtfToUniChar(src, &ch);
+        bytes = TclUtfToUniChar(src, &ch);
 	upChar = Tcl_UniCharToUpper(ch);
 
 	/*
@@ -878,7 +884,7 @@ Tcl_UtfToLower(str)
 
     src = dst = str;
     while (*src) {
-	bytes = Tcl_UtfToUniChar(src, &ch);
+	bytes = TclUtfToUniChar(src, &ch);
 	lowChar = Tcl_UniCharToLower(ch);
 
 	/*
@@ -934,7 +940,7 @@ Tcl_UtfToTitle(str)
     src = dst = str;
 
     if (*src) {
-	bytes = Tcl_UtfToUniChar(src, &ch);
+	bytes = TclUtfToUniChar(src, &ch);
 	titleChar = Tcl_UniCharToTitle(ch);
 
 	if (bytes < UtfCount(titleChar)) {
@@ -946,7 +952,7 @@ Tcl_UtfToTitle(str)
 	src += bytes;
     }
     while (*src) {
-	bytes = Tcl_UtfToUniChar(src, &ch);
+	bytes = TclUtfToUniChar(src, &ch);
 	lowChar = Tcl_UniCharToLower(ch);
 
 	if (bytes < UtfCount(lowChar)) {
@@ -1041,8 +1047,8 @@ Tcl_UtfNcmp(cs, ct, n)
 	 * This should be called only when both strings are of
 	 * at least n chars long (no need for \0 check)
 	 */
-	cs += Tcl_UtfToUniChar(cs, &ch1);
-	ct += Tcl_UtfToUniChar(ct, &ch2);
+	cs += TclUtfToUniChar(cs, &ch1);
+	ct += TclUtfToUniChar(ct, &ch2);
 	if (ch1 != ch2) {
 	    return (ch1 - ch2);
 	}
@@ -1081,8 +1087,8 @@ Tcl_UtfNcasecmp(cs, ct, n)
 	 * This should be called only when both strings are of
 	 * at least n chars long (no need for \0 check)
 	 */
-	cs += Tcl_UtfToUniChar(cs, &ch1);
-	ct += Tcl_UtfToUniChar(ct, &ch2);
+	cs += TclUtfToUniChar(cs, &ch1);
+	ct += TclUtfToUniChar(ct, &ch2);
 	if (ch1 != ch2) {
 	    ch1 = Tcl_UniCharToLower(ch1);
 	    ch2 = Tcl_UniCharToLower(ch2);
@@ -1285,9 +1291,12 @@ Tcl_UniCharNcasecmp(cs, ct, n)
     unsigned long n;			/* Number of unichars to compare. */
 {
     for ( ; n != 0; n--, cs++, ct++) {
-	if ((*cs != *ct) &&
-		(Tcl_UniCharToLower(*cs) != Tcl_UniCharToLower(*ct))) {
-	    return (*cs - *ct);
+	if (*cs != *ct) {
+	    Tcl_UniChar lcs = Tcl_UniCharToLower(*cs);
+	    Tcl_UniChar lct = Tcl_UniCharToLower(*ct);
+	    if (lcs != lct) {
+		return (lcs - lct);
+	    }
 	}
     }
     return 0;
@@ -1575,7 +1584,10 @@ Tcl_UniCharIsWordChar(ch)
  *
  *	See if a particular Unicode string matches a particular pattern.
  *	Allows case insensitivity.  This is the Unicode equivalent of
- *	the char* Tcl_StringCaseMatch.
+ *	the char* Tcl_StringCaseMatch.  The UniChar strings must be
+ *	NULL-terminated.  This has no provision for counted UniChar
+ *	strings, thus should not be used where NULLs are expected in the
+ *	UniChar string.  Use TclUniCharMatch where possible.
  *
  * Results:
  *	The return value is 1 if string matches pattern, and
@@ -1746,3 +1758,191 @@ Tcl_UniCharCaseMatch(string, pattern, nocase)
 	pattern++;
     }
 }
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * TclUniCharMatch --
+ *
+ *	See if a particular Unicode string matches a particular pattern.
+ *	Allows case insensitivity.  This is the Unicode equivalent of the
+ *	char* Tcl_StringCaseMatch.  This variant of Tcl_UniCharCaseMatch
+ *	uses counted Strings, so embedded NULLs are allowed.
+ *
+ * Results:
+ *	The return value is 1 if string matches pattern, and
+ *	0 otherwise.  The matching operation permits the following
+ *	special characters in the pattern: *?\[] (see the manual
+ *	entry for details on what these mean).
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+TclUniCharMatch(string, strLen, pattern, ptnLen, nocase)
+    CONST Tcl_UniChar *string;	/* Unicode String. */
+    int strLen;			/* length of String */
+    CONST Tcl_UniChar *pattern;	/* Pattern, which may contain special
+				 * characters. */
+    int ptnLen;			/* length of Pattern */
+    int nocase;			/* 0 for case sensitive, 1 for insensitive */
+{
+    CONST Tcl_UniChar *stringEnd, *patternEnd;
+    Tcl_UniChar p;
+
+    stringEnd  = string + strLen;
+    patternEnd = pattern + ptnLen;
+
+    while (1) {
+	/*
+	 * See if we're at the end of both the pattern and the string.  If
+	 * so, we succeeded.  If we're at the end of the pattern but not at
+	 * the end of the string, we failed.
+	 */
+
+	if (pattern == patternEnd) {
+	    return (string == stringEnd);
+	}
+	p = *pattern;
+	if ((string == stringEnd) && (p != '*')) {
+	    return 0;
+	}
+
+	/*
+	 * Check for a "*" as the next pattern character.  It matches any
+	 * substring.  We handle this by skipping all the characters up to the
+	 * next matching one in the pattern, and then calling ourselves
+	 * recursively for each postfix of string, until either we match or we
+	 * reach the end of the string.
+	 */
+
+	if (p == '*') {
+	    /*
+	     * Skip all successive *'s in the pattern
+	     */
+	    while (*(++pattern) == '*') {}
+	    if (pattern == patternEnd) {
+		return 1;
+	    }
+	    p = *pattern;
+	    if (nocase) {
+		p = Tcl_UniCharToLower(p);
+	    }
+	    while (1) {
+		/*
+		 * Optimization for matching - cruise through the string
+		 * quickly if the next char in the pattern isn't a special
+		 * character
+		 */
+		if ((p != '[') && (p != '?') && (p != '\\')) {
+		    if (nocase) {
+			while ((string < stringEnd) && (p != *string)
+				&& (p != Tcl_UniCharToLower(*string))) {
+			    string++;
+			}
+		    } else {
+			while ((string < stringEnd) && (p != *string)) {
+			    string++;
+			}
+		    }
+		}
+		if (TclUniCharMatch(string, stringEnd - string,
+			pattern, patternEnd - pattern, nocase)) {
+		    return 1;
+		}
+		if (string == stringEnd) {
+		    return 0;
+		}
+		string++;
+	    }
+	}
+
+	/*
+	 * Check for a "?" as the next pattern character.  It matches
+	 * any single character.
+	 */
+
+	if (p == '?') {
+	    pattern++;
+	    string++;
+	    continue;
+	}
+
+	/*
+	 * Check for a "[" as the next pattern character.  It is followed
+	 * by a list of characters that are acceptable, or by a range
+	 * (two characters separated by "-").
+	 */
+
+	if (p == '[') {
+	    Tcl_UniChar ch1, startChar, endChar;
+
+	    pattern++;
+	    ch1 = (nocase ? Tcl_UniCharToLower(*string) : *string);
+	    string++;
+	    while (1) {
+		if ((*pattern == ']') || (pattern == patternEnd)) {
+		    return 0;
+		}
+		startChar = (nocase ? Tcl_UniCharToLower(*pattern) : *pattern);
+		pattern++;
+		if (*pattern == '-') {
+		    pattern++;
+		    if (pattern == patternEnd) {
+			return 0;
+		    }
+		    endChar = (nocase ? Tcl_UniCharToLower(*pattern)
+			    : *pattern);
+		    pattern++;
+		    if (((startChar <= ch1) && (ch1 <= endChar))
+			    || ((endChar <= ch1) && (ch1 <= startChar))) {
+			/*
+			 * Matches ranges of form [a-z] or [z-a].
+			 */
+			break;
+		    }
+		} else if (startChar == ch1) {
+		    break;
+		}
+	    }
+	    while (*pattern != ']') {
+		if (pattern == patternEnd) {
+		    pattern--;
+		    break;
+		}
+		pattern++;
+	    }
+	    pattern++;
+	    continue;
+	}
+
+	/*
+	 * If the next pattern character is '\', just strip off the '\'
+	 * so we do exact matching on the character that follows.
+	 */
+
+	if (p == '\\') {
+	    if (++pattern == patternEnd) {
+		return 0;
+	    }
+	}
+
+	/*
+	 * There's no special character.  Just make sure that the next
+	 * bytes of each string match.
+	 */
+
+	if (nocase) {
+	    if (Tcl_UniCharToLower(*string) != Tcl_UniCharToLower(*pattern)) {
+		return 0;
+	    }
+	} else if (*string != *pattern) {
+	    return 0;
+	}
+	string++;
+	pattern++;
+    }
+}
